# -*- coding: utf-8 -*-
from adbutils.exceptions import AdbBaseError


class MinicapStartError(AdbBaseError):
    """ An error while minicap server start error """


class MinicapServerConnectError(AdbBaseError):
    """ An error while minicap server connect error """

